//============================================================
//
//  video.h - SDL implementation of MAME video routines
//
//  Copyright (c) 1996-2006, Nicola Salmoria and the MAME Team.
//  Visit http://mamedev.org for licensing and usage restrictions.
//
//  SDLMAME by Olivier Galibert and R. Belmont
//
//============================================================

#ifndef __SDL_VIDEO__
#define __SDL_VIDEO__

#include "osd_cpu.h"
#include "mamecore.h"

//============================================================
//  CONSTANTS
//============================================================

#define MAX_VIDEO_WINDOWS   	(4)

#define VIDEO_MODE_SOFT		0
#define VIDEO_MODE_OPENGL	1

#define SDL_MULTIMON	(0)

//============================================================
//  TYPE DEFINITIONS
//============================================================

typedef struct _sdl_monitor_info sdl_monitor_info;
struct _sdl_monitor_info
{
	sdl_monitor_info  *	next;					// pointer to next monitor in list
	UINT32			handle;					// handle to the monitor
	int			monitor_width;
	int			monitor_height;
	char			monitor_device[64];
	float				aspect;					// computed/configured aspect ratio of the physical device
	int					reqwidth;				// requested width for this monitor
	int					reqheight;				// requested height for this monitor
};


typedef struct _sdl_window_config sdl_window_config;
struct _sdl_window_config
{
	float				aspect;						// decoded aspect ratio
	int					width;						// decoded width
	int					height;						// decoded height
	int					depth;						// decoded depth
	int					refresh;					// decoded refresh
};


typedef struct _sdl_video_config sdl_video_config;
struct _sdl_video_config
{
	// performance options
	int					fastforward;				// fast forward?
	int					autoframeskip;				// autoframeskip?
	int					frameskip;					// explicit frameskip
	int					throttle;					// throttle speed?
	int					sleep;						// allow sleeping?
	int					novideo;				// don't draw, for pure CPU benchmarking

	// misc options
	int					framestorun;				// number of frames to run

	// global configuration
	int					windowed;	 			// start windowed?
	int					prescale;				// prescale factor (not currently supported)
	int					keepaspect;	 			// keep aspect ratio?
	int					numscreens;	 			// number of screens
	int					layerconfig;				// default configuration of layers
	int					centerh;
	int					centerv;


	// per-window configuration
	sdl_window_config	window[MAX_VIDEO_WINDOWS];		// configuration data per-window

	// hardware options
	int					mode;						// output mode
	int					waitvsync;					// spin until vsync
	int					syncrefresh;				// sync only to refresh rate
	int					triplebuf;					// triple buffer
	int					switchres;					// switch resolutions

	int					fullstretch;

	// ddraw options
	int					hwstretch;					// stretch using the hardware

	// d3d options
	int					filter;						// enable filtering

	// vector options
	float					beamwidth;				// beam width
};



//============================================================
//  GLOBAL VARIABLES
//============================================================

extern sdl_monitor_info *sdl_monitor_list;

extern int video_orientation;


extern sdl_video_config video_config;


//============================================================
//  PROTOTYPES
//============================================================

int sdlvideo_init(running_machine *machine);

void sdlvideo_monitor_refresh(sdl_monitor_info *monitor);
float sdlvideo_monitor_get_aspect(sdl_monitor_info *monitor);
sdl_monitor_info *sdlvideo_monitor_from_handle(UINT32 monitor);

void sdlvideo_set_frameskip(int frameskip);		// <0 = auto
int sdlvideo_get_frameskip(void);	   		// <0 = auto

#endif
